@@ -78,7 +78,7 @@ LibraryJSEventLoop = {
     return emSetImmediate(function() {
       {{{ runtimeKeepalivePop(); }}}
       callUserCallback(function() {
-        {{{ makeDynCall('vi', 'cb') }}}(userData);
+        {{{ makeDynCall('vp', 'cb') }}}(userData);
       });
     });
   },
@@ -90,27 +90,27 @@ LibraryJSEventLoop = {
     emClearImmediate(id);
   },
 
-  emscripten_set_immediate_loop__sig: 'vpp' ,
+  emscripten_set_immediate_loop__sig: 'ipp' ,
   emscripten_set_immediate_loop__deps: ['$polyfillSetImmediate', '$callUserCallback'],
   emscripten_set_immediate_loop: function(cb, userData) {
     polyfillSetImmediate();
     function tick() {
-      {{{ runtimeKeepalivePop(); }}}
       callUserCallback(function() {
-        if ({{{ makeDynCall('ii', 'cb') }}}(userData)) {
-          {{{ runtimeKeepalivePush(); }}}
+        if ({{{ makeDynCall('ip', 'cb') }}}(userData)) {
           emSetImmediate(tick);
+        } else {
+          {{{ runtimeKeepalivePop(); }}}
         }
       });
     }
     {{{ runtimeKeepalivePush(); }}}
-    return emSetImmediate(tick);
+    emSetImmediate(tick);
   },
 
   emscripten_set_timeout__sig: 'ipdp',
   emscripten_set_timeout__deps: ['$safeSetTimeout'],
   emscripten_set_timeout: function(cb, msecs, userData) {
-    return safeSetTimeout(() => {{{ makeDynCall('vi', 'cb') }}}(userData), msecs);
+    return safeSetTimeout(() => {{{ makeDynCall('vp', 'cb') }}}(userData), msecs);
   },
 
   emscripten_clear_timeout__sig: 'vi',
@@ -155,6 +155,61 @@ LibraryJSEventLoop = {
     {{{ runtimeKeepalivePop() }}}
     clearInterval(id);
   },
+
+  $promiseMap__deps: ['$handleAllocator'],
+  $promiseMap: "new handleAllocator();",
+
+  // Create a new promise that can be resolved or rejected by passing a unique
+  // ID to emscripten_promise_resolve/emscripten_promise_reject.  Returns a JS
+  // object containing the promise, its unique ID, and the associated
+  // reject/resolve functions.
+  $newNativePromise__deps: ['$promiseMap'],
+  $newNativePromise: function(nativeFunc, userData) {
+    var nativePromise = {};
+    var promiseId;
+    var promise = new Promise((resolve, reject) => {
+      nativePromise.reject = reject;
+      nativePromise.resolve = resolve;
+      nativePromise.id = promiseMap.allocate(nativePromise);
+#if RUNTIME_DEBUG
+      dbg('newNativePromise: ' + nativePromise.id);
+#endif
+      nativeFunc(userData, nativePromise.id);
+    });
+    nativePromise.promise = promise;
+    return nativePromise;
+  },
+
+  $getPromise__deps: ['$promiseMap'],
+  $getPromise: function(id) {
+    return promiseMap.get(id).promise;
+  },
+
+  emscripten_promise_create__sig: 'ipp',
+  emscripten_promise_create__deps: ['$newNativePromise'],
+  emscripten_promise_create: function(funcPtr, userData) {
+    return newNativePromise({{{ makeDynCall('vpi', 'funcPtr') }}}, userData).id;
+  },
+
+  emscripten_promise_resolve__deps: ['$promiseMap'],
+  emscripten_promise_resolve__sig: 'vip',
+  emscripten_promise_resolve: function(id, value) {
+#if RUNTIME_DEBUG
+    err('emscripten_resolve_promise: ' + id);
+#endif
+    promiseMap.get(id).resolve(value);
+    promiseMap.free(id);
+  },
+
+  emscripten_promise_reject__deps: ['$promiseMap'],
+  emscripten_promise_reject__sig: 'vi',
+  emscripten_promise_reject: function(id) {
+#if RUNTIME_DEBUG
+    dbg('emscripten_promise_reject: ' + id);
+#endif
+    promiseMap.get(id).reject();
+    promiseMap.free(id);
+  },
 };
 
 mergeInto(LibraryManager.library, LibraryJSEventLoop);
