@@ -8,6 +8,7 @@
 
 #include <dirent.h>
 #include <emscripten/emscripten.h>
+#include <emscripten/heap.h>
 #include <emscripten/html5.h>
 #include <errno.h>
 #include <mutex>
@@ -15,7 +16,9 @@
 #include <stdarg.h>
 #include <stdlib.h>
 #include <sys/ioctl.h>
+#include <sys/mman.h>
 #include <sys/stat.h>
+#include <sys/statfs.h>
 #include <syscall_arch.h>
 #include <unistd.h>
 #include <utility>
@@ -27,7 +30,7 @@
 #include "file_table.h"
 #include "paths.h"
 #include "pipe_backend.h"
-#include "streams.h"
+#include "special_files.h"
 #include "wasmfs.h"
 
 // File permission macros for wasmfs.
@@ -65,7 +68,7 @@ int __syscall_dup3(int oldfd, int newfd, int flags) {
 
   // If the file descriptor newfd was previously open, it will just be
   // overwritten silently.
-  fileTable.setEntry(newfd, oldOpenFile);
+  (void)fileTable.setEntry(newfd, oldOpenFile);
   return newfd;
 }
 
@@ -93,138 +96,154 @@ static __wasi_errno_t writeAtOffset(OffsetHandling setOffset,
                                     size_t iovs_len,
                                     __wasi_size_t* nwritten,
                                     __wasi_filesize_t offset = 0) {
-  if (iovs_len < 0 || offset < 0) {
-    return __WASI_ERRNO_INVAL;
-  }
-
   auto openFile = wasmFS.getFileTable().locked().getEntry(fd);
-
   if (!openFile) {
     return __WASI_ERRNO_BADF;
   }
 
-  // TODO: Check open file access mode for write permissions.
+  if (iovs_len < 0 || offset < 0) {
+    return __WASI_ERRNO_INVAL;
+  }
 
   auto lockedOpenFile = openFile->locked();
-  assert(lockedOpenFile.getFile());
   auto file = lockedOpenFile.getFile()->dynCast<DataFile>();
-
-  // If file is nullptr, then the file was not a DataFile.
-  // TODO: change to add support for symlinks.
   if (!file) {
     return __WASI_ERRNO_ISDIR;
   }
 
   auto lockedFile = file->locked();
 
-  off_t currOffset = setOffset == OffsetHandling::OpenFileState
-                       ? lockedOpenFile.getPosition()
-                       : offset;
-  off_t oldOffset = currOffset;
-  auto finish = [&] {
-    *nwritten = currOffset - oldOffset;
-    if (setOffset == OffsetHandling::OpenFileState &&
-        lockedOpenFile.getFile()->isSeekable()) {
-      lockedOpenFile.setPosition(currOffset);
+  if (setOffset == OffsetHandling::OpenFileState) {
+    if (lockedOpenFile.getFlags() & O_APPEND) {
+      off_t size = lockedFile.getSize();
+      if (size < 0) {
+        // Translate to WASI standard of positive return codes.
+        return -size;
+      }
+      offset = size;
+      lockedOpenFile.setPosition(offset);
+    } else {
+      offset = lockedOpenFile.getPosition();
     }
-  };
+  }
+
+  // TODO: Check open file access mode for write permissions.
+
+  size_t bytesWritten = 0;
   for (size_t i = 0; i < iovs_len; i++) {
     const uint8_t* buf = iovs[i].buf;
     off_t len = iovs[i].buf_len;
 
-    // Check if the sum of the buf_len values overflows an off_t (63 bits).
-    if (addWillOverFlow(currOffset, len)) {
-      return __WASI_ERRNO_FBIG;
-    }
-
     // Check if buf_len specifies a positive length buffer but buf is a
     // null pointer
     if (!buf && len > 0) {
       return __WASI_ERRNO_INVAL;
     }
 
-    auto result = lockedFile.write(buf, len, currOffset);
+    // Check if the sum of the buf_len values overflows an off_t (63 bits).
+    if (addWillOverFlow(offset, (__wasi_filesize_t)bytesWritten)) {
+      return __WASI_ERRNO_FBIG;
+    }
 
-    if (result != __WASI_ERRNO_SUCCESS) {
-      finish();
-      return result;
+    auto result = lockedFile.write(buf, len, offset + bytesWritten);
+    if (result < 0) {
+      // This individual write failed. Report the error unless we've already
+      // written some bytes, in which case report a successful short write.
+      if (bytesWritten > 0) {
+        break;
+      }
+      return -result;
+    }
+    // The write was successful.
+    bytesWritten += result;
+    if (result < len) {
+      // The read was short, so stop here.
+      break;
     }
-    currOffset += len;
   }
-  finish();
+  *nwritten = bytesWritten;
+  if (setOffset == OffsetHandling::OpenFileState &&
+      lockedOpenFile.getFile()->isSeekable()) {
+    lockedOpenFile.setPosition(offset + bytesWritten);
+  }
+  if (bytesWritten) {
+    lockedFile.setMTime(time(NULL));
+  }
   return __WASI_ERRNO_SUCCESS;
 }
 
 // Internal read function called by __wasi_fd_read and __wasi_fd_pread
 // Receives an open file state offset.
 // Optionally sets open file state offset.
+// TODO: combine this with writeAtOffset because the code is nearly identical.
 static __wasi_errno_t readAtOffset(OffsetHandling setOffset,
                                    __wasi_fd_t fd,
                                    const __wasi_iovec_t* iovs,
                                    size_t iovs_len,
                                    __wasi_size_t* nread,
                                    __wasi_filesize_t offset = 0) {
-  if (iovs_len < 0 || offset < 0) {
-    return __WASI_ERRNO_INVAL;
-  }
-
   auto openFile = wasmFS.getFileTable().locked().getEntry(fd);
-
   if (!openFile) {
     return __WASI_ERRNO_BADF;
   }
 
+  auto lockedOpenFile = openFile->locked();
+
+  if (setOffset == OffsetHandling::OpenFileState) {
+    offset = lockedOpenFile.getPosition();
+  }
+
+  if (iovs_len < 0 || offset < 0) {
+    return __WASI_ERRNO_INVAL;
+  }
+
   // TODO: Check open file access mode for read permissions.
 
-  auto lockedOpenFile = openFile->locked();
   auto file = lockedOpenFile.getFile()->dynCast<DataFile>();
 
   // If file is nullptr, then the file was not a DataFile.
-  // TODO: change to add support for symlinks.
   if (!file) {
     return __WASI_ERRNO_ISDIR;
   }
 
   auto lockedFile = file->locked();
 
-  off_t currOffset = setOffset == OffsetHandling::OpenFileState
-                       ? lockedOpenFile.getPosition()
-                       : offset;
-  off_t oldOffset = currOffset;
-  auto finish = [&] {
-    *nread = currOffset - oldOffset;
-    if (setOffset == OffsetHandling::OpenFileState &&
-        lockedOpenFile.getFile()->isSeekable()) {
-      lockedOpenFile.setPosition(currOffset);
-    }
-  };
-  size_t size = lockedFile.getSize();
+  size_t bytesRead = 0;
   for (size_t i = 0; i < iovs_len; i++) {
-    // Check if currOffset has exceeded size of file data.
-    ssize_t dataLeft = size - currOffset;
-    if (dataLeft <= 0) {
-      break;
-    }
-
     uint8_t* buf = iovs[i].buf;
+    size_t len = iovs[i].buf_len;
 
-    // Check if buf_len specifies a positive length buffer
-    // but buf is a null pointer.
-    if (!buf && iovs[i].buf_len > 0) {
+    if (!buf && len > 0) {
       return __WASI_ERRNO_INVAL;
     }
 
-    size_t bytesToRead = std::min(size_t(dataLeft), iovs[i].buf_len);
+    // TODO: Check for overflow when adding offset + bytesRead.
+    auto result = lockedFile.read(buf, len, offset + bytesRead);
+    if (result < 0) {
+      // This individual read failed. Report the error unless we've already read
+      // some bytes, in which case report a successful short read.
+      if (bytesRead > 0) {
+        break;
+      }
+      return -result;
+    }
 
-    auto result = lockedFile.read(buf, bytesToRead, currOffset);
+    // The read was successful.
 
-    if (result != __WASI_ERRNO_SUCCESS) {
-      finish();
-      return result;
+    // Backends must only return len or less.
+    assert(result <= len);
+
+    bytesRead += result;
+    if (result < len) {
+      // The read was short, so stop here.
+      break;
     }
-    currOffset += bytesToRead;
   }
-  finish();
+  *nread = bytesRead;
+  if (setOffset == OffsetHandling::OpenFileState &&
+      lockedOpenFile.getFile()->isSeekable()) {
+    lockedOpenFile.setPosition(offset + bytesRead);
+  }
   return __WASI_ERRNO_SUCCESS;
 }
 
@@ -262,11 +281,22 @@ __wasi_errno_t __wasi_fd_pread(__wasi_fd_t fd,
 }
 
 __wasi_errno_t __wasi_fd_close(__wasi_fd_t fd) {
-  auto fileTable = wasmFS.getFileTable().locked();
-  if (!fileTable.getEntry(fd)) {
-    return __WASI_ERRNO_BADF;
+  std::shared_ptr<DataFile> closee;
+  {
+    // Do not hold the file table lock while performing the close.
+    auto fileTable = wasmFS.getFileTable().locked();
+    auto entry = fileTable.getEntry(fd);
+    if (!entry) {
+      return __WASI_ERRNO_BADF;
+    }
+    closee = fileTable.setEntry(fd, nullptr);
+  }
+  if (closee) {
+    // Translate to WASI standard of positive return codes.
+    int ret = -closee->locked().close();
+    assert(ret >= 0);
+    return ret;
   }
-  fileTable.setEntry(fd, nullptr);
   return __WASI_ERRNO_SUCCESS;
 }
 
@@ -280,7 +310,10 @@ __wasi_errno_t __wasi_fd_sync(__wasi_fd_t fd) {
   // way. TODO: in the future we may want syncing of directories.
   auto dataFile = openFile->locked().getFile()->dynCast<DataFile>();
   if (dataFile) {
-    dataFile->locked().flush();
+    auto ret = dataFile->locked().flush();
+    assert(ret <= 0);
+    // Translate to WASI standard of positive return codes.
+    return -ret;
   }
 
   return __WASI_ERRNO_SUCCESS;
@@ -310,9 +343,7 @@ backend_t wasmfs_get_backend_by_path(const char* path) {
   return parsed.getFile()->getBackend();
 }
 
-static
-
-int __syscall_fstatat64(int dirfd, intptr_t path, intptr_t buf, int flags) {
+int __syscall_newfstatat(int dirfd, intptr_t path, intptr_t buf, int flags) {
   // Only accept valid flags.
   if (flags & ~(AT_EMPTY_PATH | AT_NO_AUTOMOUNT | AT_SYMLINK_NOFOLLOW)) {
     // TODO: Test this case.
@@ -328,14 +359,18 @@ int __syscall_fstatat64(int dirfd, intptr_t path, intptr_t buf, int flags) {
   auto lockedFile = file->locked();
   auto buffer = (struct stat*)buf;
 
-  buffer->st_size = lockedFile.getSize();
+  off_t size = lockedFile.getSize();
+  if (size < 0) {
+    return size;
+  }
+  buffer->st_size = size;
 
   // ATTN: hard-coded constant values are copied from the existing JS file
   // system. Specific values were chosen to match existing library_fs.js
   // values.
   // ID of device containing file: Hardcode 1 for now, no meaning at the
-  buffer->st_dev = 1;
   // moment for Emscripten.
+  buffer->st_dev = 1;
   buffer->st_mode = lockedFile.getMode();
   buffer->st_ino = file->getIno();
   // The number of hard links is 1 since they are unsupported.
@@ -343,7 +378,7 @@ int __syscall_fstatat64(int dirfd, intptr_t path, intptr_t buf, int flags) {
   buffer->st_uid = 0;
   buffer->st_gid = 0;
   // Device ID (if special file) No meaning right now for Emscripten.
-  buffer->st_rdev = 1;
+  buffer->st_rdev = 0;
   // The syscall docs state this is hardcoded to # of 512 byte blocks.
   buffer->st_blocks = (buffer->st_size + 511) / 512;
   // Specifies the preferred blocksize for efficient disk I/O.
@@ -355,15 +390,15 @@ int __syscall_fstatat64(int dirfd, intptr_t path, intptr_t buf, int flags) {
 }
 
 int __syscall_stat64(intptr_t path, intptr_t buf) {
-  return __syscall_fstatat64(AT_FDCWD, path, buf, 0);
+  return __syscall_newfstatat(AT_FDCWD, path, buf, 0);
 }
 
 int __syscall_lstat64(intptr_t path, intptr_t buf) {
-  return __syscall_fstatat64(AT_FDCWD, path, buf, AT_SYMLINK_NOFOLLOW);
+  return __syscall_newfstatat(AT_FDCWD, path, buf, AT_SYMLINK_NOFOLLOW);
 }
 
 int __syscall_fstat64(int fd, intptr_t buf) {
-  return __syscall_fstatat64(fd, (intptr_t)"", buf, AT_EMPTY_PATH);
+  return __syscall_newfstatat(fd, (intptr_t) "", buf, AT_EMPTY_PATH);
 }
 
 // When calling doOpen(), we may request an FD be returned, or we may not need
@@ -383,9 +418,9 @@ static __wasi_fd_t doOpen(path::ParsedParent parsed,
   }
 
   // TODO: remove assert when all functionality is complete.
-  assert((flags & ~(O_CREAT | O_EXCL | O_DIRECTORY | O_TRUNC | O_APPEND |
-                    O_RDWR | O_WRONLY | O_RDONLY | O_LARGEFILE | O_CLOEXEC)) ==
-         0);
+  assert((flags &
+          ~(O_CREAT | O_EXCL | O_DIRECTORY | O_TRUNC | O_APPEND | O_RDWR |
+            O_WRONLY | O_RDONLY | O_LARGEFILE | O_NOFOLLOW | O_CLOEXEC)) == 0);
 
   if (auto err = parsed.getError()) {
     return err;
@@ -425,21 +460,58 @@ static __wasi_fd_t doOpen(path::ParsedParent parsed,
       std::shared_ptr<File> created;
       if (backend == parent->getBackend()) {
         created = lockedParent.insertDataFile(std::string(childName), mode);
+        if (!created) {
+          // TODO Receive a specific error code, and report it here. For now,
+          //      report a generic error.
+          return -EIO;
+        }
       } else {
         created = backend->createFile(mode);
-        bool mounted = lockedParent.mountChild(std::string(childName), created);
+        if (!created) {
+          // TODO Receive a specific error code, and report it here. For now,
+          //      report a generic error.
+          return -EIO;
+        }
+        [[maybe_unused]] bool mounted = lockedParent.mountChild(std::string(childName), created);
         assert(mounted);
       }
       // TODO: Check that the insert actually succeeds.
       if (returnMode == OpenReturnMode::Nothing) {
         return 0;
       }
-      auto openFile = std::make_shared<OpenFileState>(0, flags, created);
+
+      std::shared_ptr<OpenFileState> openFile;
+      if (auto err = OpenFileState::create(created, flags, openFile)) {
+        assert(err < 0);
+        return err;
+      }
       return wasmFS.getFileTable().locked().addEntry(openFile);
     }
   }
 
-  // TODO: Return EISDIR if child is a directory and write access is requested.
+  if (auto link = child->dynCast<Symlink>()) {
+    if (flags & O_NOFOLLOW) {
+      return -ELOOP;
+    }
+    // TODO: The link dereference count starts back at 0 here. We could
+    // propagate it from the previous path parsing instead.
+    auto target = link->getTarget();
+    auto parsedLink = path::getFileFrom(parent, target);
+    if (auto err = parsedLink.getError()) {
+      return err;
+    }
+    child = parsedLink.getFile();
+  }
+  assert(!child->is<Symlink>());
+
+  // Return an error if the file exists and O_CREAT and O_EXCL are specified.
+  if ((flags & O_EXCL) && (flags & O_CREAT)) {
+    return -EEXIST;
+  }
+
+  if (child->is<Directory>() && accessMode != O_RDONLY) {
+    return -EISDIR;
+  }
 
   // Check user permissions.
   auto fileMode = child->locked().getMode();
@@ -457,12 +529,26 @@ static __wasi_fd_t doOpen(path::ParsedParent parsed,
     return -ENOTDIR;
   }
 
-  // Return an error if the file exists and O_CREAT and O_EXCL are specified.
-  if (flags & O_EXCL && flags & O_CREAT) {
-    return -EEXIST;
+  // Note that we open the file before truncating it because some backends may
+  // truncate opened files more efficiently (e.g. OPFS).
+  std::shared_ptr<OpenFileState> openFile;
+  if (auto err = OpenFileState::create(child, flags, openFile)) {
+    assert(err < 0);
+    return err;
+  }
+
+  // If O_TRUNC, truncate the file if possible.
+  if (flags & O_TRUNC) {
+    if (!child->is<DataFile>()) {
+      return -EISDIR;
+    }
+    if ((fileMode & WASMFS_PERM_WRITE) == 0) {
+      return -EACCES;
+    }
+    // Try to truncate the file, continuing silently if we cannot.
+    (void)child->cast<DataFile>()->locked().setSize(0);
   }
 
-  auto openFile = std::make_shared<OpenFileState>(0, flags, child);
   return wasmFS.getFileTable().locked().addEntry(openFile);
 }
 
@@ -535,24 +621,25 @@ doMkdir(path::ParsedParent parsed, int mode, backend_t backend = NullBackend) {
     backend = parent->getBackend();
   }
 
-  std::shared_ptr<File> created;
   if (backend == parent->getBackend()) {
-    created = lockedParent.insertDirectory(childName, mode);
+    if (!lockedParent.insertDirectory(childName, mode)) {
+      // TODO Receive a specific error code, and report it here. For now, report
+      //      a generic error.
+      return -EIO;
+    }
   } else {
-    created = backend->createDirectory(mode);
-    bool mounted = lockedParent.mountChild(childName, created);
+    auto created = backend->createDirectory(mode);
+    if (!created) {
+      // TODO Receive a specific error code, and report it here. For now, report
+      //      a generic error.
+      return -EIO;
+    }
+    [[maybe_unused]] bool mounted = lockedParent.mountChild(childName, created);
     assert(mounted);
   }
 
   // TODO: Check that the insertion is successful.
 
-  // Update the times.
-  auto lockedFile = created->locked();
-  time_t now = time(NULL);
-  lockedFile.setATime(now);
-  lockedFile.setMTime(now);
-  lockedFile.setCTime(now);
-
   return 0;
 }
 
@@ -591,7 +678,12 @@ __wasi_errno_t __wasi_fd_seek(__wasi_fd_t fd,
   } else if (whence == SEEK_END) {
     // Only the open file state is altered in seek. Locking the underlying
     // data file here once is sufficient.
-    position = lockedOpenFile.getFile()->locked().getSize() + offset;
+    off_t size = lockedOpenFile.getFile()->locked().getSize();
+    if (size < 0) {
+      // Translate to WASI standard of positive return codes.
+      return -size;
+    }
+    position = size + offset;
   } else {
     return __WASI_ERRNO_INVAL;
   }
@@ -609,7 +701,7 @@ __wasi_errno_t __wasi_fd_seek(__wasi_fd_t fd,
   return __WASI_ERRNO_SUCCESS;
 }
 
-int doChdir(std::shared_ptr<File>& file) {
+static int doChdir(std::shared_ptr<File>& file) {
   auto dir = file->dynCast<Directory>();
   if (!dir) {
     return -ENOTDIR;
@@ -675,7 +767,7 @@ int __syscall_getcwd(intptr_t buf, size_t size) {
   // Check if the size argument is less than the length of the absolute
   // pathname of the working directory, including null terminator.
   if (len >= size) {
-    return -ENAMETOOLONG;
+    return -ERANGE;
   }
 
   // Return value is a null-terminated c string.
@@ -759,10 +851,7 @@ int __syscall_unlinkat(int dirfd, intptr_t path, int flags) {
   }
 
   // Input is valid, perform the unlink.
-  if (!lockedParent.removeChild(childName)) {
-    return -EPERM;
-  }
-  return 0;
+  return lockedParent.removeChild(childName);
 }
 
 int __syscall_rmdir(intptr_t path) {
@@ -803,7 +892,10 @@ int __syscall_getdents64(int fd, intptr_t dirp, size_t count) {
     {".", File::DirectoryKind, dir->getIno()},
     {"..", File::DirectoryKind, parent->getIno()}};
   auto dirEntries = lockedDir.getEntries();
-  entries.insert(entries.end(), dirEntries.begin(), dirEntries.end());
+  if (int err = dirEntries.getError()) {
+    return err;
+  }
+  entries.insert(entries.end(), dirEntries->begin(), dirEntries->end());
 
   off_t bytesRead = 0;
   for (; index < entries.size() && bytesRead + sizeof(dirent) <= count;
@@ -906,7 +998,10 @@ int __syscall_renameat(int olddirfd,
     return -EACCES;
   }
 
-  // TODO: Check that the source and parent directories have the same backends.
+  // Both parents must have the same backend.
+  if (oldParent->getBackend() != newParent->getBackend()) {
+    return -EXDEV;
+  }
 
   // Check that oldDir is not an ancestor of newDir.
   for (auto curr = newParent; curr != root; curr = curr->locked().getParent()) {
@@ -936,8 +1031,9 @@ int __syscall_renameat(int olddirfd,
   }
 
   // Perform the move.
-  if (!lockedNewParent.insertMove(newFileName, oldFile)) {
-    return -EPERM;
+  if (auto err = lockedNewParent.insertMove(newFileName, oldFile)) {
+    assert(err < 0);
+    return err;
   }
   return 0;
 }
@@ -973,7 +1069,8 @@ int __syscall_symlink(intptr_t target, intptr_t linkpath) {
 
 // TODO: Test this with non-AT_FDCWD values.
 int __syscall_readlinkat(int dirfd, intptr_t path, intptr_t buf, size_t bufsize) {
-  auto parsed = path::parseFile((char*)path, dirfd);
+  // TODO: Handle empty paths.
+  auto parsed = path::parseFile((char*)path, dirfd, path::NoFollowLinks);
   if (auto err = parsed.getError()) {
     return err;
   }
@@ -1038,7 +1135,10 @@ int __syscall_fchmodat(int dirfd, intptr_t path, int mode, ...) {
   if (auto err = parsed.getError()) {
     return err;
   }
-  parsed.getFile()->locked().setMode(mode);
+  auto lockedFile = parsed.getFile()->locked();
+  lockedFile.setMode(mode);
+  // On POSIX, ctime is updated on metadata changes, like chmod.
+  lockedFile.setCTime(time(NULL));
   return 0;
 }
 
@@ -1046,6 +1146,17 @@ int __syscall_chmod(intptr_t path, int mode) {
   return __syscall_fchmodat(AT_FDCWD, path, mode, 0);
 }
 
+int __syscall_fchmod(int fd, int mode) {
+  auto openFile = wasmFS.getFileTable().locked().getEntry(fd);
+  if (!openFile) {
+    return -EBADF;
+  }
+  auto lockedFile = openFile->locked().getFile()->locked();
+  lockedFile.setMode(mode);
+  lockedFile.setCTime(time(NULL));
+  return 0;
+}
+
 int __syscall_fchownat(
   int dirfd, intptr_t path, int owner, int group, int flags) {
   // Only accept valid flags.
@@ -1103,9 +1214,9 @@ int __syscall_faccessat(int dirfd, intptr_t path, int amode, int flags) {
 
 static int doTruncate(std::shared_ptr<File>& file, off_t size) {
   auto dataFile = file->dynCast<DataFile>();
-  // TODO: support for symlinks.
+
   if (!dataFile) {
-    return __WASI_ERRNO_ISDIR;
+    return -EISDIR;
   }
 
   auto locked = dataFile->locked();
@@ -1117,11 +1228,9 @@ static int doTruncate(std::shared_ptr<File>& file, off_t size) {
     return -EINVAL;
   }
 
-  // TODO: error handling for allocation errors. atm with exceptions disabled,
-  //       however, C++ backends using std::vector for storage have no way to
-  //       report that, and will abort in malloc.
-  locked.setSize(size);
-  return 0;
+  int ret = locked.setSize(size);
+  assert(ret <= 0);
+  return ret;
 }
 
 int __syscall_truncate64(intptr_t path, uint64_t size) {
@@ -1150,9 +1259,8 @@ int __syscall_ftruncate64(int fd, uint64_t size) {
 static bool isTTY(std::shared_ptr<File>& file) {
   // TODO: Full TTY support. For now, just see stdin/out/err as terminals and
   //       nothing else.
-  return file == StdinFile::getSingleton() ||
-         file == StdoutFile::getSingleton() ||
-         file == StderrFile::getSingleton();
+  return file == SpecialFiles::getStdin() ||
+         file == SpecialFiles::getStdout() || file == SpecialFiles::getStderr();
 }
 
 int __syscall_ioctl(int fd, int request, ...) {
@@ -1178,13 +1286,8 @@ int __syscall_ioctl(int fd, int request, ...) {
       // TTY operations that we do nothing for anyhow can just be ignored.
       return -0;
     }
-    case TIOCGPGRP:
-    case TIOCSPGRP: {
-      // TODO We should get/set the group number here.
-      return -EINVAL;
-    }
     default: {
-      abort();
+      return -EINVAL; // not supported
     }
   }
 }
@@ -1200,11 +1303,13 @@ int __syscall_pipe(intptr_t fd) {
   auto reader = std::make_shared<PipeFile>(S_IRUGO, data);
   auto writer = std::make_shared<PipeFile>(S_IWUGO, data);
 
+  std::shared_ptr<OpenFileState> openReader, openWriter;
+  (void)OpenFileState::create(reader, O_RDONLY, openReader);
+  (void)OpenFileState::create(writer, O_WRONLY, openWriter);
+
   auto fileTable = wasmFS.getFileTable().locked();
-  fds[0] =
-    fileTable.addEntry(std::make_shared<OpenFileState>(0, O_RDONLY, reader));
-  fds[1] =
-    fileTable.addEntry(std::make_shared<OpenFileState>(0, O_WRONLY, writer));
+  fds[0] = fileTable.addEntry(openReader);
+  fds[1] = fileTable.addEntry(openWriter);
 
   return 0;
 }
@@ -1239,8 +1344,8 @@ int __syscall_poll(intptr_t fds_, int nfds, int timeout) {
       if (writeBit && (flags == O_RDONLY || flags == O_RDWR)) {
         // If there is data in the file, then there is also the ability to read.
         // TODO: Does this need to consider the position as well? That is, if
-        //       the position is at the end, we can't read from the current
-        //       position at least.
+        // the position is at the end, we can't read from the current position
+        // at least. If we update this, make sure the size isn't an error!
         if (openFile->locked().getFile()->locked().getSize() > 0) {
           mask |= writeBit;
         }
@@ -1284,14 +1389,19 @@ int __syscall_fallocate(int fd, int mode, uint64_t off, uint64_t len) {
     return -EINVAL;
   }
 
-  // TODO: We silently do nothing if the stream does not support allocation, but
-  //       in principle we should return EOPNOTSUPP.
   // TODO: We could only fill zeros for regions that were completely unused
   //       before, which for a backend with sparse data storage could make a
   //       difference. For that we'd need a new backend API.
   auto newNeededSize = off + len;
-  if (newNeededSize > locked.getSize()) {
-    locked.setSize(newNeededSize);
+  off_t size = locked.getSize();
+  if (size < 0) {
+    return size;
+  }
+  if (newNeededSize > size) {
+    if (auto err = locked.setSize(newNeededSize)) {
+      assert(err < 0);
+      return err;
+    }
   }
 
   return 0;
@@ -1319,7 +1429,7 @@ int __syscall_fcntl64(int fd, int cmd, ...) {
       // TODO: Should we check for a limit on the max FD number, if we have one?
       while (1) {
         if (!fileTable.getEntry(newfd)) {
-          fileTable.setEntry(newfd, openFile);
+          (void)fileTable.setEntry(newfd, openFile);
           return newfd;
         }
         newfd++;
@@ -1388,6 +1498,159 @@ int __syscall_fcntl64(int fd, int cmd, ...) {
   }
 }
 
+static int doStatFS(std::shared_ptr<File>& file, size_t size, struct statfs* buf) {
+  if (size != sizeof(struct statfs)) {
+    // We only know how to write to a standard statfs, not even a truncated one.
+    return -EINVAL;
+  }
+
+  // NOTE: None of the constants here are true. We're just returning safe and
+  //       sane values, that match the long-existing JS FS behavior (except for
+  //       the inode number, where we can do better).
+  buf->f_type = 0;
+  buf->f_bsize = 4096;
+  buf->f_frsize = 4096;
+  buf->f_blocks = 1000000;
+  buf->f_bfree = 500000;
+  buf->f_bavail = 500000;
+  buf->f_files = file->getIno();
+  buf->f_ffree = 1000000;
+  buf->f_fsid = {0, 0};
+  buf->f_flags = ST_NOSUID;
+  buf->f_namelen = 255;
+  return 0;
+}
+
+int __syscall_statfs64(intptr_t path, size_t size, intptr_t buf) {
+  auto parsed = path::parseFile((char*)path);
+  if (auto err = parsed.getError()) {
+    return err;
+  }
+  return doStatFS(parsed.getFile(), size, (struct statfs*)buf);
+}
+
+int __syscall_fstatfs64(int fd, size_t size, intptr_t buf) {
+  auto openFile = wasmFS.getFileTable().locked().getEntry(fd);
+  if (!openFile) {
+    return -EBADF;
+  }
+  return doStatFS(openFile->locked().getFile(), size, (struct statfs*)buf);
+}
+
+int _mmap_js(size_t length,
+             int prot,
+             int flags,
+             int fd,
+             size_t offset,
+             int* allocated,
+             void** addr) {
+  // PROT_EXEC is not supported (although we pretend to support the absence of
+  // PROT_READ or PROT_WRITE).
+  if ((prot & PROT_EXEC)) {
+    return -EPERM;
+  }
+
+  // One of MAP_PRIVATE, MAP_SHARED, or MAP_SHARED_VALIDATE must be used.
+  int mapType = flags & MAP_TYPE;
+  if (mapType != MAP_PRIVATE && mapType != MAP_SHARED &&
+      mapType != MAP_SHARED_VALIDATE) {
+    return -EINVAL;
+  }
+
+  if (mapType == MAP_SHARED_VALIDATE) {
+    WASMFS_UNREACHABLE("TODO: MAP_SHARED_VALIDATE");
+  }
+
+  auto openFile = wasmFS.getFileTable().locked().getEntry(fd);
+  if (!openFile) {
+    return -EBADF;
+  }
+
+  std::shared_ptr<DataFile> file;
+
+  // Keep the open file info locked only for as long as we need that.
+  {
+    auto lockedOpenFile = openFile->locked();
+
+    // Check permissions. We always need read permissions, since we need to read
+    // the data in the file to map it.
+    if ((lockedOpenFile.getFlags() & O_ACCMODE) == O_WRONLY) {
+      return -EACCES;
+    }
+
+    // According to the POSIX spec it is possible to write to a file opened in
+    // read-only mode with MAP_PRIVATE flag, as all modifications will be
+    // visible only in the memory of the current process.
+    if ((prot & PROT_WRITE) != 0 && mapType != MAP_PRIVATE &&
+        (lockedOpenFile.getFlags() & O_ACCMODE) != O_RDWR) {
+      return -EACCES;
+    }
+
+    file = lockedOpenFile.getFile()->dynCast<DataFile>();
+  }
+
+  if (!file) {
+    return -ENODEV;
+  }
+
+  // TODO: On MAP_SHARED, install the mapping on the DataFile object itself so
+  // that reads and writes can be redirected to the mapped region and so that
+  // the mapping can correctly outlive the file being closed. This will require
+  // changes to emscripten_mmap.c as well.
+
+  // Align to a wasm page size, as we expect in the future to get wasm
+  // primitives to do this work, and those would presumably be aligned to a page
+  // size. Aligning now avoids confusion later.
+  uint8_t* ptr = (uint8_t*)emscripten_builtin_memalign(WASM_PAGE_SIZE, length);
+  if (!ptr) {
+    return -ENOMEM;
+  }
+
+  auto nread = file->locked().read(ptr, length, offset);
+  if (nread < 0) {
+    // The read failed. Report the error, but first free the allocation.
+    emscripten_builtin_free(ptr);
+    return nread;
+  }
+
+  // From here on, we have succeeded, and can mark the allocation as having
+  // occurred (which means that the caller has the responsibility to free it).
+  *allocated = true;
+  *addr = (void*)ptr;
+
+  // The read must be of a valid amount, or we have had an internal logic error.
+  assert(nread <= length);
+
+  // mmap clears any extra bytes after the data itself.
+  memset(ptr + nread, 0, length - nread);
+
+  return 0;
+}
+
+int _msync_js(
+  intptr_t addr, size_t length, int prot, int flags, int fd, size_t offset) {
+  // TODO: This is not correct! Mappings should be associated with files, not
+  // fds. Only need to sync if shared and writes are allowed.
+  int mapType = flags & MAP_TYPE;
+  if (mapType == MAP_SHARED && (prot & PROT_WRITE)) {
+    __wasi_ciovec_t iovec;
+    iovec.buf = (uint8_t*)addr;
+    iovec.buf_len = length;
+    __wasi_size_t nwritten;
+    // Translate from WASI positive error codes to negative error codes.
+    return -__wasi_fd_pwrite(fd, &iovec, 1, offset, &nwritten);
+  }
+  return 0;
+}
+
+int _munmap_js(
+  intptr_t addr, size_t length, int prot, int flags, int fd, size_t offset) {
+  // TODO: This is not correct! Mappings should be associated with files, not
+  // fds.
+  // TODO: Syncing should probably be handled in __syscall_munmap instead.
+  return _msync_js(addr, length, prot, flags, fd, offset);
+}
+
 // Stubs (at least for now)
 
 int __syscall_accept4(int sockfd,
