@@ -235,6 +235,7 @@ function hasSideEffects(node) {
       case 'VariableDeclarator':
       case 'ObjectExpression':
       case 'Property':
+      case 'SpreadElement':
       case 'BlockStatement':
       case 'ArrayExpression':
       case 'EmptyStatement': {
@@ -399,7 +400,11 @@ function runJSDCE(ast, aggressive) {
       ObjectExpression(node, c) {
         // ignore the property identifiers
         node.properties.forEach(function (node) {
-          c(node.value);
+          if (node.value) {
+            c(node.value);
+          } else if (node.argument) {
+            c(node.argument);
+          }
         });
       },
       MemberExpression(node, c) {
@@ -448,18 +453,18 @@ function runAJSDCE(ast) {
   runJSDCE(ast, /* aggressive= */ true);
 }
 
-function isAsmLibraryArgAssign(node) {
-  // var asmLibraryArg = ..
+function isWasmImportsAssign(node) {
+  // var wasmImports = ..
   return (
     node.type === 'VariableDeclaration' &&
     node.declarations.length === 1 &&
-    node.declarations[0].id.name === 'asmLibraryArg' &&
+    node.declarations[0].id.name === 'wasmImports' &&
     node.declarations[0].init &&
     node.declarations[0].init.type === 'ObjectExpression'
   );
 }
 
-function getAsmLibraryArgValue(node) {
+function getWasmImportsValue(node) {
   return node.declarations[0].init;
 }
 
@@ -508,8 +513,8 @@ function isModuleAsmUse(node) {
 function applyImportAndExportNameChanges(ast) {
   const mapping = extraInfo.mapping;
   fullWalk(ast, function (node) {
-    if (isAsmLibraryArgAssign(node)) {
-      const assignedObject = getAsmLibraryArgValue(node);
+    if (isWasmImportsAssign(node)) {
+      const assignedObject = getWasmImportsValue(node);
       assignedObject.properties.forEach(function (item) {
         if (mapping[item.key.value]) {
           setLiteralValue(item.key, mapping[item.key.value]);
@@ -579,6 +584,36 @@ function isDynamicDynCall(node) {
   );
 }
 
+//
+// Matches the wasm export wrappers generated by emcc (see make_export_wrappers
+// in emscripten.py). For example:
+//
+//   var _foo = function() {
+//     return (_foo = Module['asm']['foo']).apply(null, argument)
+//   }
+//
+function isExportWrapperFunction(f) {
+  if (f.body.type != 'BlockStatement') return null;
+  if (f.body.body.length != 1) return null;
+  const expr = f.body.body[0];
+  if (expr.type == 'ReturnStatement') {
+    const rtn = expr.argument;
+    if (rtn.type == 'CallExpression') {
+      let target = rtn.callee.object;
+      if (target.type == 'ParenthesizedExpression') {
+        target = target.expression;
+      }
+      if (target.type == 'AssignmentExpression') {
+        const rhs = target.right;
+        if (isAsmUse(rhs)) {
+          return getAsmOrModuleUseName(rhs);
+        }
+      }
+    }
+  }
+  return null;
+}
+
 //
 // Emit the DCE graph, to help optimize the combined JS+wasm.
 // This finds where JS depends on wasm, and where wasm depends
@@ -626,7 +661,7 @@ function emitDCEGraph(ast) {
   //
   // The imports that wasm receives look like this:
   //
-  //  var asmLibraryArg = { "abort": abort, "assert": assert, [..] };
+  //  var wasmImports = { "abort": abort, "assert": assert, [..] };
   //
   // The exports are trickier, as they have a different form whether or not
   // async compilation is enabled. It can be either:
@@ -635,6 +670,10 @@ function emitDCEGraph(ast) {
   //
   // or
   //
+  //  var _malloc = asm["_malloc"];
+  //
+  // or
+  //
   //  var _malloc = Module["_malloc"] = (function() {
   //   return Module["asm"]["_malloc"].apply(null, arguments);
   //  });
@@ -655,7 +694,7 @@ function emitDCEGraph(ast) {
   const modulePropertyToGraphName = {};
   const exportNameToGraphName = {}; // identical to asm['..'] nameToGraphName
   const graph = [];
-  let foundAsmLibraryArgAssign = false;
+  let foundWasmImportsAssign = false;
   let foundMinimalRuntimeExports = false;
 
   function saveAsmExport(name, asmName) {
@@ -671,8 +710,8 @@ function emitDCEGraph(ast) {
   }
 
   fullWalk(ast, function (node) {
-    if (isAsmLibraryArgAssign(node)) {
-      const assignedObject = getAsmLibraryArgValue(node);
+    if (isWasmImportsAssign(node)) {
+      const assignedObject = getWasmImportsValue(node);
       assignedObject.properties.forEach(function (item) {
         let value = item.value;
         if (value.type === 'Literal' || value.type === 'FunctionExpression') {
@@ -686,14 +725,28 @@ function emitDCEGraph(ast) {
         assert(value.type === 'Identifier');
         imports.push(value.name); // the name doesn't matter, only the value which is that actual thing we are importing
       });
-      foundAsmLibraryArgAssign = true;
+      foundWasmImportsAssign = true;
       emptyOut(node); // ignore this in the second pass; this does not root
     } else if (node.type === 'VariableDeclaration') {
       if (node.declarations.length === 1) {
         const item = node.declarations[0];
         const name = item.id.name;
         const value = item.init;
-        if (value && value.type === 'AssignmentExpression') {
+        if (value && isAsmUse(value)) {
+          const asmName = getAsmOrModuleUseName(value);
+          // this is
+          //  var _x = asm['x'];
+          saveAsmExport(name, asmName);
+          emptyOut(node);
+        } else if (value && value.type === 'FunctionExpression') {
+          // this is
+          //  var x = function() { return (x = Module['asm']['x']).apply .. }
+          let asmName = isExportWrapperFunction(value);
+          if (asmName) {
+            saveAsmExport(name, asmName);
+            emptyOut(node);
+          }
+        } else if (value && value.type === 'AssignmentExpression') {
           const assigned = value.left;
           if (isModuleUse(assigned) && getAsmOrModuleUseName(assigned) === name) {
             // this is
@@ -807,8 +860,8 @@ function emitDCEGraph(ast) {
   });
   // must find the info we need
   assert(
-    foundAsmLibraryArgAssign,
-    'could not find the assigment to "asmLibraryArg". perhaps --pre-js or --post-js code moved it out of the global scope? (things like that should be done after emcc runs, as they do not need to be run through the optimizer which is the special thing about --pre-js/--post-js code)'
+    foundWasmImportsAssign,
+    'could not find the assigment to "wasmImports". perhaps --pre-js or --post-js code moved it out of the global scope? (things like that should be done after emcc runs, as they do not need to be run through the optimizer which is the special thing about --pre-js/--post-js code)'
   );
   // Read exports that were declared in extraInfo
   if (extraInfo) {
@@ -927,8 +980,8 @@ function applyDCEGraphRemovals(ast) {
   const unused = new Set(extraInfo.unused);
 
   fullWalk(ast, node => {
-    if (isAsmLibraryArgAssign(node)) {
-      const assignedObject = getAsmLibraryArgValue(node);
+    if (isWasmImportsAssign(node)) {
+      const assignedObject = getWasmImportsValue(node);
       assignedObject.properties = assignedObject.properties.filter(item => {
         const name = item.key.value;
         const value = item.value;
@@ -937,6 +990,7 @@ function applyDCEGraphRemovals(ast) {
       });
     } else if (node.type === 'AssignmentExpression') {
       // when we assign to a thing we don't need, we can just remove the assign
+      //   var x = Module['x'] = asm['x'];
       const target = node.left;
       if (isAsmUse(target) || isModuleUse(target)) {
         const name = getAsmOrModuleUseName(target);
@@ -947,6 +1001,27 @@ function applyDCEGraphRemovals(ast) {
           convertToNothingInVarInit(node);
         }
       }
+    } else if (node.type === 'VariableDeclaration') {
+      // Handle the case we declare a variable but don't assign to the module:
+      //   var x = asm['x'];
+      // and
+      //   var x = function() { return (x = asm['x']).apply(...) };
+      const init = node.declarations[0].init;
+      if (init) {
+        if (isAsmUse(init)) {
+          const name = getAsmOrModuleUseName(init);
+          const full = 'emcc$export$' + name;
+          if (unused.has(full)) {
+            convertToNothingInVarInit(init);
+          }
+        } else if (init.type == 'FunctionExpression') {
+          const name = isExportWrapperFunction(init);
+          const full = 'emcc$export$' + name;
+          if (unused.has(full)) {
+            convertToNothingInVarInit(init);
+          }
+        }
+      }
     } else if (node.type === 'ExpressionStatement') {
       const expr = node.expression;
       // In the minimal runtime code pattern we have just
@@ -1711,7 +1786,7 @@ function minifyLocals(ast) {
 function minifyGlobals(ast) {
   // The input is in form
   //
-  //   function instantiate(asmLibraryArg, wasmMemory, wasmTable) {
+  //   function instantiate(wasmImports, wasmMemory, wasmTable) {
   //      var helper..
   //      function asmFunc(global, env, buffer) {
   //        var memory = env.memory;
@@ -1725,7 +1800,7 @@ function minifyGlobals(ast) {
   // simple - as wasm2js output is - and looks at all the minifiable names as
   // a whole. A possible bug here is something like
   //
-  //   function instantiate(asmLibraryArg, wasmMemory, wasmTable) {
+  //   function instantiate(wasmImports, wasmMemory, wasmTable) {
   //      var x = foo;
   //      function asmFunc(global, env, buffer) {
   //        var foo = 10;
